/*----------------------------------------------------------------------
  File    : cluster.h
  Contents: cluster and cluster set management
  Author  : Christian Borgelt
  History : 2001.08.17 file created
            2001.09.05 cluster size added
            2001.09.11 functions cls_reg() added
            2001.11.08 functions cls_center(), cls_covmat() etc. added
            2002.08.07 parsing function completed
            2003.01.30 data normalization variables added
            2003.03.11 weight parameter added to function cls_reg()
            2003.05.15 functions combined into function cls_factors()
            2003.05.15 membership normalization modes added
            2003.05.18 cls_norm made a real function, cls_isovar() added
            2003.06.08 parameter radius added to function cls_type()
            2003.08.11 adapted to new module attmap
            2003.08.12 adapted to new module nstats
            2003.08.15 adapted to new module radfn
            2004.02.27 function cls_deletex added
            2004.02.28 size and shape regularization added
            2004.03.18 normalization parameters added
            2004.04.23 field CLUSTER.det removed (var is sufficient)
            2004.07.13 definition of CLS_UNIT added
            2004.07.28 parameter 'conly' added to function cls_update
            2004.08.12 adapted to new module parse
            2004.08.18 backpropagation functions added
            2007.01.10 use of attribute maps changed
            2007.11.26 noise cluster configured with separate function
            2008.02.29 redesign for feature weighting completed
            2013.08.16 adapted to definition of type DIMID in matrix.h
            2018.06.10 other initialization methods added
            2018.08.18 extended functions (attset/table) removed
            2018.08.27 annular algorithm [Drake 2013] added
            2018.08.28 yinyang algorithm [Ding et al. 2015] added
            2018.08.29 stratum algorithm (modified annular) added
            2018.08.31 shallot algorithm (improved exponion) added
            2018.09.01 coordinate grid (stratum) algorithm added
            2019.07.09 functions cls_ctrjtr() added (jitter coordinates)
            2019.07.10 non-squared distance based kmeans++ init. added
----------------------------------------------------------------------*/
#ifndef __CLUSTER__
#define __CLUSTER__
#include <stdio.h>
#include <math.h>
#include "fntypes.h"
#include "radfn.h"
#ifdef CLS_PARSE
#ifndef NST_PARSE
#define NST_PARSE
#endif
#endif
#include "nstats.h"
#include "matrix.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- prototype flags --- */
#define CLS_CENTER       0x00   /* use cluster center (mandatory) */
#define CLS_SIZE         0x01   /* use cluster size */
#define CLS_VARS         0x02   /* use variances (independent dims.) */
#define CLS_COVARS       0x04   /* use covariance matrix */
#define CLS_JOINT        0x08   /* use joint (co)variances */
#define CLS_WEIGHT       0x10   /* use relative cluster weight */
#define CLS_NORM         0x20   /* normalize membership function */
#define CLS_ALL          0x3f   /* all prototype flags */

/* --- initialization modes --- */
/*      CLS_CENTER          0 *//* center of the data space */
#define CLS_UNIFORM         1   /* samples from uniform distribution */
#define CLS_DIAG            2   /* diagonal of data space */
#define CLS_LATIN           3   /* latin hypercube sampling */
#define CLS_POINTS          4   /* points chosen uniformly at random */
#define CLS_MAXIMIN         5   /* chosen by maximum minimum distance */
#define CLS_KMPP            6   /* (full)    kmeans++ (d^2 distrib.) */
#define CLS_QKMPP           7   /* quantiled kmeans++ (d^2 distrib.) */
#define CLS_AFKMC2          8   /* Markov chain of d^2 distribution */
#define CLS_SAFKMC2         9   /* simplified Markov chain of d^2 d. */
#define CLS_RKMPP          10   /* (full)    kmeans++ (d distrib.) */
#define CLS_RQKMPP         11   /* quantiled kmeans++ (d distrib.) */
#define CLS_RAFKMC2        12   /* Markov chain of d distribution */
#define CLS_RSAFKMC2       13   /* simplified Markov chain of d dist. */

/* --- normalization modes --- */
#define CLS_NONE            0   /* no normalization */
#define CLS_SUM1            1   /* normalize to sum 1 */
#define CLS_MAX1            2   /* normalize to maximum 1 */
#define CLS_HARD            3   /* hard assignment (winner takes all) */

/* --- update methods --- */
#define CLS_GRADIENT     0x00   /* gradient descent/ascent */
#define CLS_ALTOPT       0x01   /* alternating optimization */
#define CLS_COMPLRN      0x02   /* competitive learning */
#define CLS_RBFGRAD      0x03   /* gradient descent for RBFs */
#define CLS_STANDARD     0x10   /* standard plain center update */
#define CLS_ELKAN        0x11   /* method from [Elkan   2003] */
#define CLS_HAMERLY      0x12   /* method from [Hamerly 2010] */
#define CLS_SHALLOT      0x13   /* shallot/peral onion algorithm */
#define CLS_ANNULAR      0x14   /* annular algorithm [Drake 2013] */
#define CLS_STRATUM      0x15   /* stratum algorithm */
#define CLS_COGRID       0x16   /* coordinate grid algorithm */
#define CLS_HYBRID       0x17   /* hybrid of Hamerly and Elkan */
#define CLS_HYSHALL      0x18   /* hybrid plus shallot algorithm */
#define CLS_YINYANG      0x19   /* yinyang algorithm [Ding e.a. 2015] */
#define CLS_PLAIN        0x10   /* plain (center) update method */
#define CLS_METHOD       0xff   /* mask for update method */
#define CLS_EXPONION     (CLS_HAMERLY|CLS_C2CALL)

/* --- update modifiers --- */
#define CLS_EXPAND     0x0100   /* expand change by a factor */
#define CLS_MOMENTUM   0x0200   /* update with momentum term */
#define CLS_ADAPTIVE   0x0300   /* self-adaptive change factor */
#define CLS_RESILIENT  0x0400   /* resilient update */
#define CLS_QUICK      0x0500   /* quickprop analog */
#define CLS_C2CDIST    0x0800   /* cluster to cluster distances */
#define CLS_C2CMIN     0x0900   /* only minimum distances */
#define CLS_C2CALL     0x0a00   /* all pairwise distances */
#define CLS_MODIFIER   0x0f00   /* mask for update modifier */

/* --- extended update modifiers --- */
#define CLS_UNIT       0x1000   /* unit length center vectors */
#define CLS_MODVAR     0x2000   /* use modifier also for (co)vars. */
#define CLS_EXTMOD     0xf000   /* mask for extended update modifier */

/* --- validity measures --- */
#define CLS_OBJFN           0   /* objective function */
#define CLS_COVERAGE        1   /* coverage of data points */
#define CLS_FDBIDX          2   /* fuzzy Davies-Bouldin index */
#define CLS_PCOEFF          3   /* partition coefficient */
#define CLS_PCNORM          4   /* normalized partition coefficient */
#define CLS_PENTROPY        5   /* partition entropy */
#define CLS_FHYPVOL         6   /* fuzzy hyper-volume */
#define CLS_PRPEXP          7   /* proportion exponent */
#define CLS_XBISEP          8   /* Xie-Beni index / separation */
#define CLS_PARTDENS        9   /* partition density */
#define CLS_PDAVG          10   /* average partition density */
#define CLS_NOISE          11   /* membership sum to noise cluster */
#define CLS_VMCNT          12   /* number of validity measures */

/* --- description modes --- */
#define CLS_TITLE      0x0001   /* print a title (as a comment) */
#define CLS_INFO       0x0002   /* print a title (as a comment) */
#define CLS_RBFNET     0x0004   /* describe as part of RBF network */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- a cluster --- */
  double  *ctr;                 /* cluster center */
  MATRIX  *cov;                 /* covariance matrix */
  double  *dif;                 /* difference vector to center */
  MATRIX  *inv;                 /* inverse covariance matrix */
  double  *sum;                 /* weighted sum of data vectors */
  MATRIX  *smp;                 /* weighted sum of matrix products */
  double  *chc;                 /* change of center coordinates */
  MATRIX  *chv;                 /* change of (co)variances */
  double  *grc;                 /* gradient of center coordinates */
  MATRIX  *grv;                 /* gradient of (co)variances */
  double  *buf;                 /* buffer for computations */
  MATRIX  *mat;                 /* ditto */
  int     ok;                   /* whether update will be ok */
  int     dec;                  /* flag for computed decomposition */
  double  var;                  /* equivalent isotropic variance */
  double  size;                 /* size (usually equal to var) */
  double  scale;                /* scaling factor for update */
  double  d2;                   /* squared distance of data point */
  double  msd;                  /* degree of membership of data point */
  double  move;                 /* how much cluster moved by update */
  double  val[4];               /* validity measure terms */
} CLUSTER;                      /* (cluster) */

typedef struct {                /* --- cluster evaluation data --- */
  int     flags;                /* flags for configured measures */
  double  rad;                  /* radius for partition density */
  double  cnt;                  /* number of processed patterns */
  double  sum;                  /* sum of membership degrees */
  double  sqr;                  /* sum of squared membership degrees */
  double  ent;                  /* entropy sum of membership degrees */
  double  pxp;                  /* proportion exponent */
  double  noise;                /* membership sum to noise cluster */
} CLSEVAL;                      /* (cluster evaluation data) */

typedef struct {                /* --- a set of clusters --- */
  int     type;                 /* type of cluster prototypes */
  int     tnew;                 /* new cluster type (before setup) */
  DIMID   inpcnt;               /* number of input dimensions */
  DIMID   clscnt;               /* number of clusters */
  DIMID   init;                 /* number of initialized clusters */
  CLUSTER *cls;                 /* vector of clusters */
  RADFN   *radfn;               /* radial function (Cauchy or Gauss) */
  DERIVFN *drvfn;               /* derivative of radial function */
  double  rfnps[2];             /* parameters of the radial function */
  double  unit;                 /* norm. factor for unit integral */
  double  noise;                /* noise cluster parameter */
  double  ncmsd[2];             /* membership degree to noise cluster */
  int     norm;                 /* membership normalization mode */
  double  nrmps[2];             /* membership normalization params. */
  double  fwexp;                /* feature weight exponent */
  double  msexp;                /* membership exponent for adaptation */
  int     setup;                /* flag for completed setup */
  int     method;               /* parameter update method + modifier */
  int     eigen;                /* flag for eigenvalue decomposition */
  int     rbfgrad;              /* flag for gradient descent for RBFs */
  double  rates[3];             /* learning rates */
  double  decay[3];             /* learning rate decay parameters */
  double  moment;               /* momentum term */
  double  growth;               /* growth    factor */
  double  shrink;               /* shrinkage factor */
  double  maxfac;               /* maximal factor (quick backprop) */
  double  minchg;               /* minimal change/learning rate */
  double  maxchg;               /* maximal change/learning rate */
  double  regps[5];             /* regularization parameters */
  int     steps;                /* number of update steps */
  int     resized;              /* flag for resized clusters */
  NSTATS  *nst;                 /* normalization statistics */
  double  *vec;                 /* buffer for a data vector */
  double  *buf;                 /* buffer (e.g. for eigenvalues) */
  MATRIX  *mat;                 /* buffer for matrix computations */
  CLSEVAL eval;                 /* cluster evaluation data */
} CLSET;                        /* (cluster set) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern CLSET*  cls_create  (DIMID dimcnt, DIMID clscnt);
extern void    cls_delete  (CLSET *clset);
extern CLSET*  cls_clone   (CLSET *clset);
extern DIMID   cls_inpcnt  (CLSET *clset);
extern DIMID   cls_dim     (CLSET *clset);
extern DIMID   cls_clscnt  (CLSET *clset);
extern DIMID   cls_cnt     (CLSET *clset);

extern NSTATS* cls_nstats  (CLSET *clset);
extern void    cls_reg     (CLSET *clset, const double *vec,
                            double weight);
extern void    cls_value   (CLSET *clset, DIMID index, double value);

extern void    cls_type    (CLSET *clset, int type, double size);
extern void    cls_radfn   (CLSET *clset, RADFN *radfn, DERIVFN *drvfn,
                            const double *params);
extern void    cls_fwexp   (CLSET *clset, double fwexp);
extern void    cls_noise   (CLSET *clset, double noise);
extern void    cls_norm    (CLSET *clset, int mode,
                            const double *params);

extern void    cls_init    (CLSET *clset, int mode,     RANDFN randfn,
                            const double *vec);
extern void    cls_jitter  (CLSET *clset, double range, RANDFN randfn);
extern void    cls_method  (CLSET *clset, int    method);
extern void    cls_msexp   (CLSET *clset, double msexp);
extern void    cls_moment  (CLSET *clset, double moment);
extern void    cls_factors (CLSET *clset, double growth, double shrink);
extern void    cls_limits  (CLSET *clset, double minchg, double maxchg);
extern void    cls_reset   (CLSET *clset);
extern void    cls_lrate   (CLSET *clset, const double *rates,
                                          const double *decay);
extern void    cls_regular (CLSET *clset, const double *params);

extern int     cls_setup   (CLSET *clset);
extern double  cls_sqrdist (CLSET *clset, DIMID clsid,
                            const double *vec);
extern double  cls_dist    (CLSET *clset, DIMID clsid,
                            const double *vec);
extern DIMID   cls_exec    (CLSET *clset, const double *vec,
                            double *msdegs);
extern void    cls_agg1    (CLSET *clset, DIMID clsid,
                            const double *vec, double weight);
extern DIMID   cls_aggr    (CLSET *clset, const double *vec,
                            double weight);
extern void    cls_bkprop  (CLSET *clset, const double *errs);
extern double  cls_uplain  (CLSET *clset);
extern double  cls_update  (CLSET *clset, int conly);
extern void    cls_sort    (CLSET *clset);

extern void    cls_clsdel  (CLSET *clset, DIMID index);
extern void    cls_clsmrg  (CLSET *clset, DIMID idx1, DIMID idx2);

extern double* cls_center  (CLSET *clset, DIMID cid);
extern MATRIX* cls_covmat  (CLSET *clset, DIMID cid);
extern MATRIX* cls_invmat  (CLSET *clset, DIMID cid);
extern double  cls_isovar  (CLSET *clset, DIMID cid);
extern double  cls_weight  (CLSET *clset, DIMID cid);
extern double  cls_msdeg   (CLSET *clset, DIMID cid);
extern double  cls_move    (CLSET *clset, DIMID cid);

extern void    cls_evcfg   (CLSET *clset, int measure, double param);
extern void    cls_evaggr  (CLSET *clset, const double *vec,
                            double weight);
extern double  cls_eval    (CLSET *clset, int measure);

extern void    cls_sclcov  (CLSET *clset);
extern void    cls_unscale (CLSET *clset);

extern int     cls_desc    (CLSET *clset,
                            FILE *file, int mode, int maxlen);
extern int     cls_cdesc   (MATRIX *ctrs, NSTATS *nst,
                            FILE *file, int mode, int maxlen);

#ifdef CLS_PARSE
extern CLSET*  cls_parse   (SCANNER *scan, DIMID inpcnt, int hdr);
extern MATRIX* cls_cparse  (SCANNER *scan, DIMID inpcnt, int hdr,
                            NSTATS **pnst);
#endif

extern int     cls_ctrsel  (MATRIX *ctrs, MATRIX *data, int method,
                            double size, double trim,
                            DIMID *outids, DIMID outcnt, RANDFN randfn);
extern void    cls_ctrjtr  (MATRIX *ctrs, double range, RANDFN randfn);
extern DIMID   cls_xhcm    (MATRIX *ctrs, MATRIX *data, int method,
                            DIMID epochs, DIMID *outids, DIMID outcnt,
                            double size, DIMID *cids);
extern double  cls_hobjfn  (MATRIX *ctrs, MATRIX *data);

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define cls_inpcnt(s)      ((s)->inpcnt)
#define cls_dim(s)         ((s)->inpcnt)
#define cls_clscnt(s)      ((s)->clscnt)
#define cls_cnt(s)         ((s)->clscnt)

#define cls_nstats(s)      ((s)->nst)
#define cls_reg(s,v,w)     nst_reg((s)->nst, v, w)

#define cls_noise(s,n)     ((s)->noise  = (n))
#define cls_fwexp(s,e)     ((s)->fwexp  = (e))
#define cls_moment(s,m)    ((s)->moment = (m))
#define cls_reset(s)       ((s)->steps  = 0)

#define cls_dist(s,i,v)    sqrt(cls_sqrdist(s,i,v))

#define cls_center(s,i)    ((s)->cls[i].ctr)
#define cls_covmat(s,i)    ((s)->cls[i].cov)
#define cls_invmat(s,i)    ((s)->cls[i].inv)
#define cls_isovar(s,i)    ((s)->cls[i].var)
#define cls_size(s,i)      ((s)->cls[i].size)
#define cls_scale(s,i)     ((s)->cls[i].scale)
#define cls_weight(s,i)    mat_getwgt((s)->cls[i].cov)
#define cls_msdeg(s,i)     ((s)->cls[i].msd)
#define cls_move(s,i)      ((s)->cls[i].move)

#endif
