/*----------------------------------------------------------------------
  File    : rbf.c
  Contents: radial basis function network management
  Author  : Christian Borgelt
  History : 2004.08.13 file created
            2004.08.15 update functions made a separate file
            2007.01.10 usage of attribute maps changed
            2013.08.19 adapted to definition of type DIMID in matrix.h
            2018.10.15 extended functions (attset, attmap) removed
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <assert.h>
#include "rbf.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE       16        /* block size for parsing */

/* error codes 0 to -15 defined in scanner.h */
#define E_ATTEXP    (-16)       /* attribute expected */
#define E_UNKATT    (-17)       /* unknown attribute */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#ifdef RBF_PARSE
static const char *msgs[] = {   /* error messages */
  /*      0 to  -7 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /*     -8 to -15 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /* E_ATTEXP  -16 */  "#attribute expected instead of '%s'",
  /* E_UNKATT  -17 */  "#unknown attribute '%s'",
};
#endif

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef void UPDATEFN (RBFNET* rbf);  /* a weight update function */

/*----------------------------------------------------------------------
  Weight Update Functions
----------------------------------------------------------------------*/

static void standard (RBFNET *rbf)
{                               /* --- standard backpropagation */
  DIMID  i;                     /* loop variable */
  double *w, *c, *g;            /* to traverse the vectors */

  w = rbf->wgts;                /* get the necessary vectors and */
  g = rbf->grds;                /* traverse the connection weights */
  if      (rbf->moment > 0) {   /* if backpropagation with momentum */
    c = rbf->chgs;              /* get the vector of old changes */
    for (i = 0; i < rbf->wgtcnt; i++) {
      w[i] += c[i] = rbf->moment *c[i] - rbf->lrate *g[i];
      g[i]  = 0;                /* update the connection weights */
    } }                         /* and reinitialize the gradient */
  else if (rbf->moment < 0) {   /* if Nesterov's accelerated gradient */
    c = rbf->chgs;              /* get the vector of old gradients */
    for (i = 0; i < rbf->wgtcnt; i++) {
      c[i]  = -rbf->moment *c[i] +g[i];
      w[i] += rbf->lrate *(rbf->moment *c[i] -g[i]);
      g[i]  = 0;                /* update the connection weights */
    } }                         /* and reinitialize the gradient */
  else {                        /* if standard backpropagation */
    for (i = 0; i < rbf->wgtcnt; i++) {
      w[i] -= rbf->lrate *g[i]; /* update the connection weights */
      g[i]  = 0;                /* and reinitialize the gradient */
    }
  }
}  /* standard() */

/*--------------------------------------------------------------------*/

static void manhattan (RBFNET *rbf)
{                               /* --- Manhattan training */
  DIMID  i;                     /* loop variable */
  double *w, *g;                /* to traverse the vectors */

  w = rbf->wgts;                /* get the necessary vectors and */
  g = rbf->grds;                /* traverse the connection weights */
  for (i = rbf->wgtcnt; --i >= 0; ) {
    if      (g[i] > 0) w[i] -= rbf->lrate;
    else if (g[i] < 0) w[i] += rbf->lrate;
    g[i] = 0;                   /* update the connection weights */
  }                             /* and clear the gradient */
}  /* manhattan() */            /* for the next step */

/*--------------------------------------------------------------------*/

static void adaptive (RBFNET *rbf)
{                               /* --- super self-adaptive backprop. */
  DIMID  i;                     /* loop variable */
  double *w, *c, *g, *p;        /* to traverse the vectors */
  double t;                     /* temporary buffer */

  w = rbf->wgts; c = rbf->chgs; /* get the necessary vectors and */
  g = rbf->grds; p = rbf->bufs; /* traverse the connection weights */
  for (i = rbf->wgtcnt; --i >= 0; ) {
    if      (g[i] > 0) t =  p[i];
    else if (g[i] < 0) t = -p[i];
    else               t =  0;  /* check directions of the changes */
    if      (t > 0) {           /* if gradients have the same sign, */
      c[i] *= rbf->growth;      /* increase the learning rate */
      if (c[i] > rbf->maxchg) c[i] = rbf->maxchg;
      p[i] = g[i]; }            /* note the current gradient */
    else if (t < 0) {           /* if gradients have opposite signs, */
      c[i] *= rbf->shrink;      /* decrease the learning rate */
      if (c[i] < rbf->minchg) c[i] = rbf->minchg;
      p[i] = 0; }               /* suppress a change in the next step */
    else {                      /* if one gradient is zero, */
      p[i] = g[i]; }            /* only note the current gradient */
    w[i] -= c[i] *g[i];         /* update the connection weight */
    g[i] = 0;                   /* and clear the gradient */
  }                             /* for the next step */
}  /* adaptive() */

/*--------------------------------------------------------------------*/

static void resilient (RBFNET *rbf)
{                               /* --- resilient backpropagation */
  DIMID  i;                     /* loop variable */
  double *w, *c, *g, *p;        /* to traverse the vectors */
  double t;                     /* temporary buffer */

  w = rbf->wgts; c = rbf->chgs; /* get the necessary vectors and */
  g = rbf->grds; p = rbf->bufs; /* traverse the connection weights */
  for (i = rbf->wgtcnt; --i >= 0; ) {
    if      (g[i] > 0) t =  p[i];
    else if (g[i] < 0) t = -p[i];
    else               t =  0;  /* check directions of the changes */
    if      (t > 0) {           /* if gradients have the same sign, */
      c[i] *= rbf->growth;      /* increase the update value */
      if (c[i] > rbf->maxchg) c[i] = rbf->maxchg;
      p[i] = g[i]; }            /* note the current gradient */
    else if (t < 0) {           /* if gradients have opposite signs, */
      c[i] *= rbf->shrink;      /* decrease the update value */
      if (c[i] < rbf->minchg) c[i] = rbf->minchg;
      p[i] = 0; }               /* suppress a change in the next step */
    else {                      /* if one gradient is zero, */
      p[i] = g[i]; }            /* only note the current gradient */
    if      (g[i] > 0) w[i] -= c[i];  /* update the weight */
    else if (g[i] < 0) w[i] += c[i];  /* (but use only the sign) */
    g[i] = 0;                   /* clear the gradient */
  }                             /* for the next step */
}  /* resilient() */

/*--------------------------------------------------------------------*/

static void quick (RBFNET *rbf)
{                               /* --- quick backpropagation */
  DIMID  i;                     /* loop variable */
  double *w, *c, *g, *p;        /* to traverse the vectors */
  double m;                     /* maximal fraction of new derivative */
  double t;                     /* temporary buffer */

  m = rbf->growth /(rbf->growth +1);
  w = rbf->wgts; c = rbf->chgs; /* get the necessary vectors and */
  g = rbf->grds; p = rbf->bufs; /* traverse the connection weights */
  for (i = rbf->wgtcnt; --i >= 0; ) {
    t = p[i] -g[i];             /* compute the change of the gradient */
    if      (t * c[i] >= 0)     /* if the parabola opens downwards, */
      c[i] = -rbf->lrate *g[i]; /* do a normal backpropagation step */
    else if (p[i] > 0) {        /* if previous gradient was positive */
      if (g[i] >= m *p[i])      /* if the growth factor would become */
        c[i] *= rbf->growth;    /* too large, use the maximal factor */
      else                      /* otherwise compute the factor */
        c[i] *= g[i] /t;        /* for a jump to the minimum */
      if (g[i] > 0)             /* if steps are in the same dir., */
        c[i] -= rbf->lrate *g[i]; } /* add a normal backprop. step */
    else if (p[i] < 0) {        /* if previous gradient was negative */
      if (g[i] <= m *p[i])      /* if the growth factor would become */
        c[i] *= rbf->growth;    /* too large, use the maximal factor */
      else                      /* otherwise compute the factor */
        c[i] *= g[i] /t;        /* for a jump to the minimum */
      if (g[i] < 0)             /* if steps are in the same dir., */
        c[i] -= rbf->lrate *g[i]; } /* add a normal backprop. step */
    else                        /* if this is the first update, */
      c[i] = -rbf->lrate *g[i]; /* do a normal backpropagation step */
    if      (c[i] >  rbf->maxchg) c[i] =  rbf->maxchg;  /* clamp the */
    else if (c[i] < -rbf->maxchg) c[i] = -rbf->maxchg;  /* change */
    w[i] += c[i];               /* adapt the connection weight, */
    p[i] = g[i]; g[i] = 0;      /* note the gradient and clear it */
  }                             /* for the next step */
}  /* quick() */

/*--------------------------------------------------------------------*/

static UPDATEFN *updatefn[] = {
  /* RBF_STANDARD   0 */  standard,
  /* RBF_MANHATTAN  1 */  manhattan,
  /* RBF_ADPATIVE   2 */  adaptive,
  /* RBF_RESILIENT  3 */  resilient,
  /* RBF_QUICK      4 */  quick,
};                              /* list of weight update functions */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

static RBFNET* create (CLSET *clset)
{                               /* --- create a base RBF network */
  DIMID  cnt;                   /* number of weights per output unit */
  RBFNET *rbf;                  /* created radial basis function net */

  assert(clset);                /* check the function arguments */
  cnt = cls_clscnt(clset);      /* get the number of hidden units */
  rbf = (RBFNET*)malloc(sizeof(RBFNET)
                      +(size_t)(cnt+cnt+1) *sizeof(double));
  if (!rbf) return NULL;        /* create a base RBF network */
  rbf->clset  = clset;          /* and initialize some fields */
  rbf->inpcnt = cls_inpcnt(clset);
  rbf->hidcnt = cnt;
  rbf->bpes   = rbf->hids +cnt+1;
  rbf->wgts   = NULL;
  rbf->mat    = mat_create(cnt+1, cnt+1);
  if (!rbf->mat) { free(rbf); return NULL; }
  return rbf;                   /* return the created network */
}  /* create() */

/*--------------------------------------------------------------------*/

static int init (RBFNET *rbf, DIMID outcnt)
{                               /* --- initialize some variables */
  size_t z;                     /* number of parameters */

  assert(rbf && (outcnt > 0));  /* check the function arguments */
  rbf->outcnt = outcnt;         /* compute number of weights */
  rbf->wgtcnt = outcnt *(rbf->hidcnt +1);
  z = (size_t)(3 *outcnt +4 *rbf->wgtcnt);
  rbf->wgts = (double*)realloc(rbf->wgts, z *sizeof(double));
  if (!rbf->wgts) return -1;    /* create parameters arrays */
  rbf->chgs = rbf->wgts +rbf->wgtcnt;/* and organize the memory */
  rbf->grds = rbf->chgs +rbf->wgtcnt;
  rbf->bufs = rbf->grds +rbf->wgtcnt;
  rbf->outs = rbf->bufs +rbf->wgtcnt;
  rbf->trgs = rbf->outs +rbf->outcnt;
  rbf->errs = rbf->trgs +rbf->outcnt;
  rbf->hids[rbf->hidcnt] = 1;   /* set fixed input for bias value */
  return 0;                     /* return 'ok' */
}  /* init() */

/*--------------------------------------------------------------------*/

RBFNET* rbf_create (CLSET *clset, DIMID outcnt)
{                               /* --- create an RBF network */
  RBFNET *rbf;                  /* created radial basis function net */

  assert(clset && (outcnt > 0));/* check the function arguments */
  rbf = create(clset);          /* create a base RBF network */
  if (!rbf) return NULL;        /* and initialize it */
  if (init(rbf, outcnt) != 0) { rbf_delete(rbf, 0); return NULL; }
  rbf_clear(rbf);               /* clear all connection weights */
  return rbf;                   /* return the created RBF network */
}  /* rbf_create() */

/*--------------------------------------------------------------------*/

void rbf_delete (RBFNET *rbf, int delcs)
{                               /* --- delete an RBF network */
  CLSET *clset;                 /* buffer for cluster set */

  assert(rbf);                  /* check the function arguments */
  clset = rbf->clset;           /* note the underlying cluster set */
  if (rbf->wgts) free(rbf->wgts);
  mat_delete(rbf->mat);         /* delete con. weights, init. matrix, */
  free(rbf);                    /* delete the base structure and */
  if (delcs) cls_delete(clset); /* the underlying cluster set */
}  /* rbf_delete() */

/*--------------------------------------------------------------------*/

void rbf_reg (RBFNET *rbf, const double *inps,
              const double *trgs, double weight)
{                               /* --- register a training pattern */
  DIMID  i, k;                  /* loop variables */
  double *w, *h, t;             /* to traverse the vectors, buffer */

  assert(rbf && rbf->mat);      /* check the function arguments */
  cls_exec(rbf->clset, inps, rbf->hids);  /* execute the cluster set */
  mat_addmp(rbf->mat, rbf->hids, weight); /* and aggregate the result */
  if (!trgs) trgs = rbf->trgs;  /* get the target values, */
  h = rbf->hids;                /* the hidden neuron outputs, */
  w = rbf->wgts +rbf->wgtcnt;   /* and the weight vector */
  for (i = rbf->outcnt; --i >= 0; ) {
    *--w += t = weight*trgs[i]; /* traverse the output units */
    for (k = rbf->hidcnt; --k >= 0; )
      *--w += h[k] *t;          /* aggregate the rhs terms */
  }                             /* for a multilinear regression */
}  /* rbf_reg() */

/*--------------------------------------------------------------------*/

void rbf_clear (RBFNET *rbf)
{                               /* --- clear the connection weights */
  assert(rbf && rbf->mat);      /* check the function argument */
  mat_init(rbf->mat, MAT_ZERO|MAT_VECTOR|MAT_WEIGHT, NULL);
  memset(rbf->wgts, 0, (size_t)rbf->wgtcnt *sizeof(double));
}  /* rbf_clear() */            /* clear init. matrix, con. weights */

/*--------------------------------------------------------------------*/

int rbf_init (RBFNET *rbf)
{                               /* --- initialize connection weights */
  DIMID  i;                     /* loop variable */
  double *w;                    /* to traverse the weights */

  assert(rbf && rbf->mat);      /* check the function argument */
  if (mat_chdecom(rbf->mat, rbf->mat) != 0)
    return -1;                  /* decompose the coefficient matrix */
  w = rbf->wgts +rbf->wgtcnt;   /* get the weights vector */
  for (i = rbf->outcnt; --i >= 0; ) {
    w -= rbf->hidcnt +1;        /* traverse the output neurons */
    mat_chsubst(w, rbf->mat,w); /* and solve the corresponding */
  }                             /* linear equation systems */
  return 0;                     /* return 'ok' */
}  /* rbf_init() */

/*--------------------------------------------------------------------*/

void rbf_setup (RBFNET *rbf)
{                               /* --- set up network for training */
  DIMID i;                      /* loop variable */

  assert(rbf);                  /* check the function argument */
  cls_norm (rbf->clset, CLS_NONE, NULL);
  cls_fwexp(rbf->clset, 0);     /* remove special features */
  cls_setup(rbf->clset);        /* and set up the cluster set */
  if ((rbf->method == RBF_RESILIENT)
  ||  (rbf->method == RBF_ADAPTIVE)) {
    for (i = 0; i < rbf->wgtcnt; i++)
      rbf->chgs[i] = rbf->lrate; }    /* initialize weight changes */
  else                          /* and clear gradients and buffers */
    memset(rbf->chgs, 0, (size_t)rbf->wgtcnt *sizeof(double));
  memset(rbf->grds, 0, (size_t)rbf->wgtcnt *sizeof(double));
  memset(rbf->bufs, 0, (size_t)rbf->wgtcnt *sizeof(double));
}  /* rbf_setup() */

/*--------------------------------------------------------------------*/

void rbf_exec (RBFNET *rbf, const double *inps, double *outs)
{                               /* --- execute an RBF network */
  DIMID  i, k;                  /* loop variables */
  double *w;                    /* to traverse the weights */

  assert(rbf);                  /* check the function arguments */
  cls_exec(rbf->clset, inps, rbf->hids);
  w = rbf->wgts +rbf->wgtcnt;   /* execute the cluster set */
  for (i = rbf->outcnt; --i >= 0; ) {
    rbf->outs[i] = *--w;        /* traverse the output neurons */
    for (k = rbf->hidcnt; --k >= 0; )
      rbf->outs[i] += *--w * rbf->hids[k];
  }                             /* compute weighted sum of outputs */
  if (outs)                     /* copy outputs to result vector */
    memcpy(outs, rbf->outs, (size_t)rbf->outcnt *sizeof(double));
}  /* rbf_exec() */

/*--------------------------------------------------------------------*/

double rbf_error (RBFNET *rbf, const double *trgs)
{                               /* --- compute sum of squared errors */
  DIMID  k;                     /* loop variable */
  double sse = 0, d;            /* sum of squared errors, buffer */

  assert(rbf);                  /* check the function arguments */
  if (!trgs) trgs = rbf->trgs;  /* if not target given, use internals */
  for (k = 0; k < rbf->outcnt; k++) {
    rbf->errs[k] = d = trgs[k] -rbf->outs[k];
    sse += d*d;                 /* compute the output errors */
  }                             /* (target - computed output) */
  return sse;                   /* return sum of squared errors */
}  /* rbf_error() */

/*--------------------------------------------------------------------*/

double rbf_bkprop (RBFNET *rbf, const double *trgs)
{                               /* --- back propagate errors */
  DIMID  i, k;                  /* loop variables */
  double *g, *w, *b;            /* to access the vectors */
  double sse, t;                /* sum of squared errors, buffer */

  assert(rbf);                  /* check the function arguments */
  sse = rbf_error(rbf, trgs);   /* compute sum of squared errors */
  b = memset(rbf->bpes, 0, (size_t)rbf->hidcnt *sizeof(double));
  g = rbf->grds +rbf->wgtcnt;   /* get the gradient vector */
  w = rbf->wgts +rbf->wgtcnt;   /* and the weight   vector */
  for (i = rbf->outcnt; --i >= 0; ) {
    *--g -= t = rbf->errs[i];   /* traverse the output errors */
    --w;                        /* skip the offset weight */
    for (k = rbf->hidcnt; --k >= 0; ) {
      *--g -= rbf->hids[k] *t;  /* update the connection weights */
      b[k] += *--w *t;          /* with the output errors */
    }                           /* and backpropagate the errors, */
  }                             /* then backpropagate to cluster set */
  cls_bkprop(rbf->clset, rbf->bpes);
  return sse;                   /* return sum of squared errors */
}  /* rbf_bkprop() */

/*--------------------------------------------------------------------*/

void rbf_update (RBFNET *rbf)
{                               /* --- update the parameters */
  DIMID i;                      /* loop variable */

  assert(rbf);                  /* check the function argument */
  if (rbf->decay != 1.0)        /* if weight decay is requested */
    for (i = 0; i < rbf->wgtcnt; i++)
      rbf->wgts[i] *= rbf->decay;  /* reduce all connection weights */
  updatefn[rbf->method](rbf);   /* call the weight update function */
  cls_update(rbf->clset, 0);    /* and update the cluster set */
}  /* rbf_update() */

/*--------------------------------------------------------------------*/

int rbf_desc (RBFNET *rbf, FILE *file, int mode, int maxlen)
{                               /* --- describe an RBF network */
  DIMID  i, k;                  /* loop variables */
  int    len, l;                /* loop variables for comments */
  double *w;                    /* to traverse the weights */
  char   *indent = "";          /* indentation string */

  assert(rbf && file);          /* check the function arguments */
  len = (maxlen > 0) ? maxlen -2 : 70;

  /* --- print header (as a comment) --- */
  if (mode & RBF_TITLE) {       /* if the title flag is set */
    fputs("/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fprintf(file, "\n  radial basis function network\n");
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* print a title header */
  }                             /* (as a comment) */
  if (maxlen <= 0) maxlen = INT_MAX;

  fputs("rbfnet = {\n", file);  /* start the neural network desc. */
  indent = "  ";                /* indent the description itself */

  /* --- describe cluster set --- */
  cls_desc(rbf->clset, file, (indent[0]) ? CLS_RBFNET : 0, maxlen);

  /* --- print connection weights --- */
  fprintf(file, "%sweights  = {", indent);
  w = rbf->wgts;                /* traverse the weight vectors */
  for (i = 0; i < rbf->outcnt; i++) {
    if (i > 0) fprintf(file, ",\n%s            ", indent);
    fputs("{ ", file);          /* print the connection weights */
    for (k = 0; k < rbf->hidcnt; k++)
      fprintf(file, "%+.16g, ", *w++);
    fprintf(file, "%+.16g }", *w++);
  }                             /* print bias value and term. list */
  fputs("};\n", file);          /* terminate the weight description */

  fputs("};\n", file);          /* terminate the network description */

  /* --- print additional information (as a comment) --- */
  if (mode & RBF_INFO) {        /* if the add. info. flag is set */
    fputs("\n/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fprintf(file, "\n  number of inputs : %"DIMID_FMT, rbf->inpcnt);
    fprintf(file, "\n  number of outputs: %"DIMID_FMT, rbf->outcnt);
    fprintf(file, "\n  number of rbfs   : %"DIMID_FMT, rbf->hidcnt);
    fprintf(file, "\n  number of weights: %"DIMID_FMT, rbf->wgtcnt);
    fputc('\n', file);          /* print additional network info. */
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* (as a comment) */
  }

  return ferror(file) ? -1 : 0; /* return the write status */
}  /* rbf_desc() */

/*--------------------------------------------------------------------*/
#ifdef RBF_PARSE

static int weights (RBFNET *rbf, SCANNER *scan, DIMID outcnt)
{                               /* --- parse the connection weights */
  DIMID  i, k, n;               /* loop variables */
  double *w, *p;                /* to traverse the weights */

  assert(rbf && scan);          /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for 'weights' */
  ||  (strcmp(scn_value(scan), "weights") != 0))
    SCN_ERROR(scan, E_STREXP, "weights");
  SCN_NEXT(scan);               /* consume 'weights' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  w = rbf->wgts; n = 0;         /* get the weight vector and */
  for (k = 0; (outcnt <= 0) || (k < outcnt); k++) {
    if (k > 0) { SCN_CHAR(scan, ','); } /* consume ',' */
    SCN_CHAR(scan, '{');        /* consume '{' */
    if (k >= n) {               /* if the weight vector is full */
      n += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
      p = (double*)realloc(rbf->wgts, (size_t)n *(size_t)(rbf->hidcnt+1)
                                      *sizeof(double));
      if (!p) SCN_ERROR(scan, E_NOMEM);
      rbf->wgts = p;            /* enlarge con. weight vector(s) */
      w = p +k *(rbf->hidcnt+1);/* and get the new weight position */
    }
    for (i = rbf->hidcnt +1; --i >= 0; ) {
      if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
      *w++ = strtod(scn_value(scan), NULL);
      SCN_NEXT(scan);           /* get the connection weight */
      if (i > 0) { SCN_CHAR(scan, ','); }
    }                           /* if not bias value, consume ',' */
    SCN_CHAR(scan, '}');        /* consume '}' */
    if ((outcnt <= 0) && (scn_token(scan) != ',')) {
      k++; break; }             /* check for more connection weights */
  }
  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  rbf->outcnt = k;              /* set the number of output units */
  return 0;                     /* return 'ok' */
}  /* weights() */

/*--------------------------------------------------------------------*/

static RBFNET* parse (SCANNER *scan, CLSET *clset, DIMID outcnt)
{                               /* --- parse an RBF network desc. */
  RBFNET *rbf;                  /* parsed radial basis function net */
  CLSET  *pc = NULL;            /* parsed cluster set */

  assert(scan);                 /* check the function arguments */
  scn_first(scan);              /* set messages, get first token */
  if (!clset) {                 /* if no cluster set is given */
    clset = pc = cls_parse(scan, 0, 0);
    if (!pc) return NULL;       /* parse the cluster set */
  }
  rbf = create(clset);          /* create a base RBF network */
  if (!rbf) { if (pc) cls_delete(pc); return NULL; }
  if ((weights(rbf, scan, outcnt) != 0) /* read connection weights */
  ||  (init(rbf, rbf->outcnt)     != 0)) {    /* and init. network */
    rbf_delete(rbf, (pc) ? 1 : 0); return NULL; }
  return rbf;                   /* return the parsed network */
}  /* parse() */

/*--------------------------------------------------------------------*/

static int header (SCANNER *scan)
{                               /* --- parse header */
  assert(scan);                 /* check the function argument */
  if ((scn_token(scan) != T_ID)
  ||  ((strcmp(scn_value(scan), "rbfnet") != 0)
  &&   (strcmp(scn_value(scan), "rbf")    != 0)))
    SCN_ERROR(scan, E_STREXP, "rbfnet");
  SCN_NEXT(scan);               /* consume 'lvqnet' or 'clset' */
  if (scn_token(scan) == '(') { /* if a target attribute follows */
    SCN_NEXT(scan);             /* consume '(' */
    if (scn_token(scan) != T_ID) SCN_ERROR(scan, E_ATTEXP);
    SCN_NEXT(scan);             /* consume the target attribute */
    SCN_CHAR(scan, ')');        /* check for and consume ')' */
  }
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  return 0;                     /* return 'ok' */
}  /* header() */

/*--------------------------------------------------------------------*/

static int trailer (SCANNER *scan)
{                               /* --- parse trailer */
  assert(scan);                 /* check the function argument */
  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* trailer() */

/*--------------------------------------------------------------------*/

RBFNET* rbf_parse (SCANNER *scan, CLSET *clset, DIMID outcnt, int hdr)
{                               /* --- parse an RBF network */
  RBFNET *rbf;                  /* parsed radial basis function net */

  assert(scan);                 /* check the function argument */
  scn_setmsgs(scan, msgs, (int)(sizeof(msgs)/sizeof(*msgs)));
  scn_first(scan);              /* set messages, get first token */
  if (hdr && (header(scan) != 0))
    return NULL;                /* parse the header if necessary */
  rbf = parse(scan, clset, outcnt);
  if (!rbf) return NULL;        /* parse the neural network desc. */
  if (hdr && (trailer(scan) != 0)) {      /* and then the trailer */
    rbf_delete(rbf, 0); return NULL; }    /* if necessary */
  return rbf;                   /* return the created RBF network */
}  /* rbf_parse() */

#endif  /* #ifdef RBF_PARSE */
